#include<iostream>
#include"utils.h"

using namespace std;

#ifdef _Cuda
extern "C" {
	void brillo_cuda(unsigned char*,unsigned char*,unsigned int,unsigned int,unsigned int,float,float);
	void cuda_clean();
}
#endif

#ifdef _Asm
extern "C" void brillo(IplImage* in,IplImage* out);
#else
void brillo(IplImage* in,IplImage* out){
	unsigned char dark = 0x00;
	unsigned char white = 0xff;
	float a,b;
	
	#define brillo(pix) (((pix).r + (pix).g + (pix).b)/3)
	for(int irow = 0; irow < in->height;++irow){
		pixel* row = (pixel*)(in->imageData + in->widthStep * irow);
		for(int icol = 0;icol < in->width;++icol){
			unsigned char br = brillo(row[icol]);
			
			dark = max(dark,br);
			white = min(white,br);			
		}
	}
	
	float w = white;
	float d = dark;
	b = (white==0)?0.0:(255.0/((-d/w)+1));
	a = (white==0)?(255.0/d):(-b/w); 

	#define f(x) (a * (x) + b)
	cout << "white: " << (int)white << endl;
	cout << "dark: " << (int)dark << endl;
	cout << "f(x) = " << a << " x + " << b << endl;

#ifdef _Cuda
	brillo_cuda((unsigned char*)in->imageData,(unsigned char*)out->imageData,in->widthStep,in->height,in->width,a,b);
#else
	for(int irow = 0; irow < in->height;++irow){
		pixel* inrow = (pixel*)(in->imageData + in->widthStep * irow);
		pixel* outrow = (pixel*)(out->imageData + out->widthStep * irow);
		for(int icol = 0;icol < in->width;++icol){
			unsigned int cr,cg,cb;
			unsigned char bn,bv; //brillo nuevo y brillo viejo
			
			bv = brillo(inrow[icol]);
			bn = f(bv);

			if(bv){
				cr = (bn * inrow[icol].r) / bv;
				cg = (bn * inrow[icol].g) / bv;
				cb = (bn * inrow[icol].b) / bv;
			}else{
				cr = cg = cb = 0;
			}

			outrow[icol].r = (cr>255)?255:cr;
			outrow[icol].g = (cg>255)?255:cg;
			outrow[icol].b = (cb>255)?255:cb;
		}
	}
#endif

	#undef f	
	#undef brillo
}
#endif

#ifdef _Video
int main(int argc,char** argv){
	if(argc<2)return 0;
	if(!strcmp(argv[1],"--help")){
		cout << "./brilloVideo [video]" << endl; 
		return 0;
	}
	bool isinit = false;
	IplImage  *frame;
	IplImage  *frameAlpha;
	IplImage  *salida;
	int       key;

	CvCapture *capture = cvCaptureFromAVI( argv[1] );
	if( !capture ){
		cout <<  "No se pudo abrir el archivo " << argv[1] << endl;
		return -1;
	}    

	int fps = (int)cvGetCaptureProperty( capture, CV_CAP_PROP_FPS );

	cvNamedWindow( "Video Original", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "Video Salida", CV_WINDOW_AUTOSIZE );

	while(1){
		frame = cvQueryFrame( capture );
            
	        if( !frame ){
		cvSetCaptureProperty(capture, CV_CAP_PROP_POS_AVI_RATIO, 0);
	        	frame = cvQueryFrame( capture );
	        }
	        if( !frame ) break;

		if(!isinit){
#ifdef _Asm	        
			frameAlpha = agregarCanalAlpha(frame);
			salida = chequearImagen(frameAlpha);
#else
			salida = chequearImagen(frame);
#endif
			isinit = true;
		}

#ifdef _Asm	        
		reusarCanalAlfa(frame,frameAlpha);
		frame = frameAlpha;
#endif
		brillo(frame, salida);
		cvShowImage( "Video Original", frame );
		cvShowImage( "Video Salida", salida );

		key = cvWaitKey( 1000 / fps ) & 0xFF;
		if(key == 27) break;
	}
	cvReleaseCapture(&capture);
#ifdef _Asm
	cvReleaseImage(&frameAlpha);
#endif
	cvReleaseImage(&salida);
	limpiarVentanas();
#ifdef _Cuda
	cuda_clean();
#endif
	return 0;
}
#else
int main(int argc,char** argv){
	if(argc<2)return 0;
	if(!strcmp(argv[1],"--help")){
		cout << "./brillo [imagein]" << endl; 
		cout << "./brillo [imagein] [imageout]" << endl; 
		return 0;
	}

	const char* imagename = argv[1]; 
	const char* outname = (argc<3)?0:argv[2];
	
	IplImage* img = cvLoadImage(imagename);
	IplImage* res = 0;
	ERROR(img,"No se pudo abrir la imagen");
#ifdef _Asm
	img = agregarCanalAlpha(img);
	res = chequearImagen(img);
#endif
	res = chequearImagen(img);
	brillo(img,res);

	if(!outname){
		mostrarVentanas(img,res);
		limpiarVentanas();
	}else{
		cvSaveImage(outname,res);
		cout << "Imagen guardada como " << outname << endl;
	}
	limpiarImagenes(img,res);
#ifdef _Cuda
	cuda_clean();
#endif
	return 0;
}
#endif
