#include<iostream>
#include"utils.h"

using namespace std;
#ifdef _Asm
extern "C" void gaussSimple(IplImage* in, IplImage* out);
#endif

#ifdef _Cuda
extern "C" {
	void gaussSimple(unsigned char*,unsigned char*,unsigned int,unsigned int,unsigned int);
	void cuda_clean();
}
#endif

#ifdef _Cpp
void gaussSimple(IplImage* in, IplImage* out)
{
	pixel* dataOut = (pixel*)out->imageData;	
	pixel* dataIn = (pixel*)in->imageData;	
	
	pixel* p_inrow_up; //Puntero a la fila anterior a la que estoy leyendo.
	pixel* p_inrow; //Puntero a la fila que estoy leyendo.
	pixel* p_inrow_down; //Puntero a la fila siguiente a la que estoy leyendo.
	pixel* p_outrow; //Puntero a la fila que estoy escribiendo.
	int filas = in->height; //Cantidad de filas.
	int cols = in->width; //Cantidad de columnas.
	int widthStep = in->widthStep; //Cantidad de bytes por fila.	
	unsigned int promR;
	unsigned int promB;
	unsigned int promG;
	
	for(int irow = 1;irow < filas-1;irow++){
		p_inrow_up = (pixel*)((int)dataIn + widthStep * (irow-1)); //calculo cual es la proxima columna
		p_inrow = (pixel*)((int)dataIn + widthStep * irow); //calculo cual es la proxima columna
		p_inrow_down = (pixel*)((int)dataIn + widthStep * (irow+1)); //calculo cual es la proxima columna
		p_outrow = (pixel*)((int)dataOut + widthStep * irow); 

		for(int icol = 1;icol < cols-1;icol++){
			promR = (p_inrow_up[icol-1].r + p_inrow_up[icol].r +  p_inrow_up[icol+1].r + p_inrow[icol-1].r + p_inrow[icol].r +  p_inrow[icol+1].r + p_inrow_down[icol-1].r + p_inrow_down[icol].r +  p_inrow_down[icol+1].r)/9;
			promB = (p_inrow_up[icol-1].b + p_inrow_up[icol].b +  p_inrow_up[icol+1].b + p_inrow[icol-1].b + p_inrow[icol].b +  p_inrow[icol+1].b + p_inrow_down[icol-1].b + p_inrow_down[icol].b +  p_inrow_down[icol+1].b)/9;
			promG = (p_inrow_up[icol-1].g + p_inrow_up[icol].g +  p_inrow_up[icol+1].g + p_inrow[icol-1].g + p_inrow[icol].g +  p_inrow[icol+1].g + p_inrow_down[icol-1].g + p_inrow_down[icol].g +  p_inrow_down[icol+1].g)/9;
			p_outrow[icol].b = promB; 
			p_outrow[icol].g = promG; 
			p_outrow[icol].r = promR;
		}
	}
	
	
	
}
#endif

#ifdef _Video
int main(int argc,char** argv){
	if(argc<2)return 0;
	if(!strcmp(argv[1],"--help")){
		cout << "./gauss [video]" << endl; 
		return 0;
	}
	bool isinit = false;
	IplImage  *frame;
	IplImage  *frameAlpha;
	IplImage  *salida;
	int       key;

	CvCapture *capture = cvCaptureFromAVI( argv[1] );
	if( !capture ){
		cout <<  "No se pudo abrir el archivo " << argv[1] << endl;
		return -1;
	}    

	int fps = (int)cvGetCaptureProperty( capture, CV_CAP_PROP_FPS );

	cvNamedWindow( "Video Original", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "Video Salida", CV_WINDOW_AUTOSIZE );

	
	unsigned int frames = 0;
	medir();
	while(1){
		frame = cvQueryFrame( capture );
            
	        if( !frame ){
			cvSetCaptureProperty(capture, CV_CAP_PROP_POS_AVI_RATIO, 0);
	        	frame = cvQueryFrame( capture );
	        }
	        if( !frame ) break;

		if(!isinit){
#ifdef _Asm	        
			frameAlpha = agregarCanalAlpha(frame);
			salida = chequearImagen(frameAlpha);
#else
			salida = chequearImagen(frame);
#endif
			isinit = true;
		}

#ifdef _Asm	        
		reusarCanalAlfa(frame,frameAlpha);
		gaussSimple(frameAlpha, salida);
#endif

#ifdef _Cpp
		gaussSimple(frame, salida);
#endif
#ifdef _Cuda
		gaussSimple((unsigned char*)frame->imageData,(unsigned char*)salida->imageData,frame->widthStep,frame->height,frame->width);
#endif
		cvShowImage( "Video Original", frame );
		cvShowImage( "Video Salida", salida );
		
		frames++;
		key = cvWaitKey( 1000 / fps ) & 0xFF;
		if(key == 27) break;
	}
	cout << "FPS: " << frames / medir() << endl;
	cvReleaseCapture(&capture);
#ifdef _Asm
	cvReleaseImage(&frameAlpha);
#endif
	cvReleaseImage(&salida);
	limpiarVentanas();
#ifdef _Cuda
	cuda_clean();
#endif
	return 0;
}
#else
int main(int argc,char** argv){
	if(argc<2)return 0;
	if(!strcmp(argv[1],"--help")){
		cout << "./gauss [imagein]" << endl; 
		cout << "./gauss [imagein] [imageout]" << endl; 
		return 0;
	}

	const char* imagename = argv[1]; 
	const char* outname = (argc<3)?0:argv[2];
	
	IplImage* img = cvLoadImage(imagename);
	IplImage* res = 0;

	ERROR(img,"No se pudo abrir la imagen");
	

#ifdef _Asm
	IplImage* conalfa = agregarCanalAlpha(img); 
	cvReleaseImage(&img);
	img = conalfa;
	ERROR(img,"No se pudo agregar el canal de alfa");
#endif

	res = chequearImagen(img);
#ifdef _Cuda
	gaussSimple((unsigned char*)img->imageData,(unsigned char*)res->imageData,img->widthStep,img->height,img->width);
#else
	gaussSimple(img,res);
#endif

	if(!outname){
		mostrarVentanas(img,res);
		limpiarVentanas();
	}else{
		cvSaveImage(outname,res);
		cout << "Imagen guardada como " << outname << endl;
	}
	limpiarImagenes(img,res);
#ifdef _Cuda
	cuda_clean();
#endif
	return 0;
}
#endif
