#include<stdio.h>
#define pixel_size 3

__global__ void cu_greyscale(unsigned char *cu_in,unsigned int in_step,unsigned char *cu_out,unsigned int out_step,unsigned int height,unsigned int width) {
	unsigned int n_pixel = blockIdx.x * blockDim.x + threadIdx.x; //me da en que numero de pixel estoy!
	if(n_pixel < height * width) {
		unsigned int acum = 0;
		unsigned int n_row,n_col;
		unsigned char* pix_in,*pix_out;

		n_row = n_pixel / width;
		n_col = n_pixel % width;

		pix_in = cu_in + n_row * in_step + pixel_size * sizeof(unsigned char) * n_col;
		pix_out = cu_out + n_row * out_step + sizeof(unsigned char) * n_col;

		acum = 0.3*(float)pix_in[0] + 0.59*(float)pix_in[1] + 0.11*(float)pix_in[2];

		*pix_out = acum;
	}
}

unsigned char *cu_mem_in = 0, *cu_mem_out = 0;
bool cu_mem_init = false;

extern "C" void greyscale(unsigned char* in_data,unsigned int in_step,unsigned char* out_data,unsigned int out_step,unsigned int height,unsigned int width) {
	if(!cu_mem_init){
		cudaMalloc((void**)&cu_mem_in,sizeof(unsigned char) * in_step * height);
		cudaMalloc((void**)&cu_mem_out,sizeof(unsigned char) * out_step * height);
		cu_mem_init = true;
	}

	cudaMemcpy(cu_mem_in,in_data,sizeof(unsigned char) * in_step * height,cudaMemcpyHostToDevice);
	dim3 dimBlock(16*16,1);
	dim3 dimGrid((width * height)/dimBlock.x + 1,1); //proceso de a tiras.

	cu_greyscale<<<dimGrid , dimBlock>>>(cu_mem_in,in_step,cu_mem_out,out_step,height,width);

	cudaMemcpy(out_data,cu_mem_out,sizeof(unsigned char) * out_step * height,cudaMemcpyDeviceToHost);
}

extern "C" void cuda_clean() {
	if(cu_mem_init) {
		cudaFree(cu_mem_in);
		cudaFree(cu_mem_out);
	}
}
