%include "macros.mac"

%define in ebp+8
%define out ebp+12

%define dataOut ebp-4
%define dataIn ebp-8
%define filas ebp-12
%define cols ebp-16
%define widthStepIn ebp-20
%define widthStepOut ebp-24
%define prow ebp-28
%define row ebp-32
%define nrow ebp-36
%define outrow ebp-40
%define noutrow ebp-44
%define nnoutrow ebp-48

%define tamStack 48

;eax -> 
;ebx -> 
;ecx -> 
;edx -> 
;esi -> irow
;edi -> icol

global zoomInterpoladorASM

section .text

zoomInterpoladorASM:
	convencionC tamStack

	mov eax, [in]
	mov eax, [eax+68]
	mov [dataIn], eax ;tengo en dataIn in->imageData
	mov eax, [out]
	mov ecx, [eax+68]
	mov [dataOut], ecx ;tengo en dataOut  out->imageData
	mov ecx, [eax+72]
	mov [widthStepOut], ecx ; tengo en widthStepOut out->widthStep
	mov eax, [in]
	mov ecx, [eax+40]
	mov [cols], ecx	;cols = in->width
	mov ecx, [eax+44]
	mov [filas], ecx ;filas = in->height
	mov ecx, [eax+72]
	mov [widthStepIn], ecx ;widthStepIn = in->widthStep

	jmp copiarOriginal
primerPasada: jmp primerosPix
segundaPasada: jmp segundosPix
termine:
	finConvencionC tamStack
	ret


copiarOriginal:
	push DWORD [row]
	push DWORD [outrow]
	xor esi, esi ;irow = 0
	.cicloFilas:
		cmp esi, [filas]
		jge .termineFilas

		mov eax, [widthStepIn]
		mul esi
		add eax, [dataIn]
		mov [row], eax ;in->imageData + in->widthStep * irow
		
		mov eax, [widthStepOut]
		add eax, eax
		mul esi
		add eax, [dataOut] 
		mov [outrow], eax ;out->imageData + out->widthStep*irow*2

		xor edi, edi
		.cicloCols:
			cmp edi, [cols]
			jge .termineCols ;icol < in->width
			
			mov eax, [cols]
			sub eax, edi
			cmp eax, 4
			jl .menosDe4

			mov eax, [row]
			movups xmm0, [eax]
			pshufd xmm1, xmm0, 01110010b ;pixeles 1 y 0
			pshufd xmm0, xmm0, 11011000b ;pixeles 4 y 3

			;dejo en negro los pixeles intermedios
			xor eax, eax
			dec eax
			movd xmm2, eax
			pshufd xmm2, xmm2, 11001100b
			pand xmm0, xmm2
			pand xmm1, xmm2

			mov eax, [outrow]
			movups [eax], xmm0
			add eax, 16
			movups [eax], xmm1
			add DWORD [row], 16
			add DWORD [outrow], 32
			jmp .seguir
		.menosDe4:
			mov ecx, eax
			.ciclo:
				mov eax, [row]
				mov ebx, [outrow]
				mov eax, [eax]
				mov [ebx], eax
				add DWORD[row],4
				add DWORD[outrow], 8
				loop .ciclo
		.seguir:
			add edi, 4 ;cols += 4
			jmp .cicloCols
		.termineCols:
			inc esi ;irow++
			jmp .cicloFilas
	.termineFilas:
		pop DWORD [outrow]
		pop DWORD [row]
		jmp primerPasada


primerosPix:
	push DWORD [row]
	push DWORD [nrow]
	push DWORD [outrow]
	
	xor esi, esi ;irow = 0
	.cicloFilas:
		cmp esi, [filas]
		jge .termineFilas

		mov eax, [widthStepIn]
		mul esi
		add eax, [dataIn]
		mov [row], eax ;in->imageData + in->widthStep * irow

		add eax, [widthStepIn]
		mov [nrow], eax ;(int)row + in->widthStep		

		mov eax, esi
		add eax, esi
		inc eax
		mul DWORD [widthStepOut]
		add eax, [dataOut] 
		mov [outrow], eax ;(pixel*)(out->imageData + out->widthStep*(irow*2 + 1))

		xor edi, edi
		.cicloCols:
			cmp edi, [cols]
			jge .termineCols ;icol < in->width

			mov eax, [cols]
			sub eax, edi
			cmp eax, 4
			jl .menosDe5 ;icol < in->width-4
			
			mov eax, [filas]
			dec eax
			cmp esi, eax
			jge .ultimaFila ;irow < in->height-1	
			
			;icol < in->width-4 && irow < in->height-1
			mov eax, [row]
			movups xmm0, [eax]
			mov eax, [nrow]
			movups xmm1, [eax]	
			pshufd xmm0, xmm0, 00011011b
			pshufd xmm1, xmm1, 00011011b
			;obtener el pixel			
			%rep 3
			movups xmm2, xmm0
			pxor xmm3, xmm3
			punpckhbw xmm2, xmm3 ;unpackeo para sumar y despues dividir
			movups xmm3, xmm1			
			pxor xmm4, xmm4
			punpckhbw xmm3, xmm4
			paddw xmm2, xmm3 ;tengo la suma de row[col;col+1] y nrow[col;col+1]
			movups xmm3, xmm2
			pshufd xmm3, xmm3, 01001111b
			paddw xmm2, xmm3 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm3, xmm3
			punpckhwd xmm2, xmm3 ;tengo en xmm2 en cada dw r,g,b,a
			cvtdq2ps xmm2, xmm2 ;tengo en xmm2 los valores en flotante
			mov eax, 4
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			cvtdq2ps xmm3, xmm3 ;tengo los 4 en flotante
			divps xmm2, xmm3; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm2, xmm2 ;lo paso a entero de nuevo
			
			;pack AMD
			pxor xmm3, xmm3
			pcmpgtd xmm3, xmm2 ;0 > xmm2?
			mov eax, -1
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b	
			pxor xmm3, xmm4			
			pand xmm2, xmm3 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			pand xmm2, xmm3 ;paso cada canal a byte
			pslld xmm2, 24	
			movups xmm3, xmm2 ;copio
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm2, xmm2, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte

			;termine un pixel
			mov eax, edi
			add eax, eax
			inc eax
			mov ebx, 4	
			mul ebx		
			add eax, [outrow]
			movd [eax], xmm2 ;guardo el pixel
			inc edi ;icol++
			pshufd xmm0, xmm0, 10010011b
			pshufd xmm1, xmm1, 10010011b
			%endrep
			add DWORD [row], 12
			add DWORD [nrow], 12

			jmp .seguir

		.menosDe5:
			mov eax, [filas]
			dec eax
			cmp esi, eax
			jge .ultimaFilaMenosDe5 ;irow < in->height-1	
			
			;icol >= in->width-4 && irow < in->height-1
			mov eax, [cols]
			sub eax, edi
			cmp eax, 1
			jz .ultimaCol
			
			dec eax
			mov ecx, eax
			.traerDeADos:
				pxor xmm0, xmm0
				pxor xmm1, xmm1
				mov eax, [row]
				;pinsrd AMD				
				movd xmm1, [eax+4]
				por xmm0, xmm1
				pxor xmm1, xmm1				
				movd xmm1, [eax]
				pshufd xmm0, xmm0, 11110011b
				por xmm0, xmm1
				;pinsrb xmm0, [eax], 0
				;inc eax
				;pinsrd xmm0, [eax], 1
				mov eax, [nrow]
				;pinsrd AMD				
				pxor xmm2, xmm2				
				movd xmm2, [eax+4]
				por xmm1, xmm2
				pxor xmm2, xmm2				
				movd xmm2, [eax]
				pshufd xmm1, xmm1, 11110011b
				por xmm1, xmm2
				;pinsrd xmm1, [eax], 0
				;inc eax
				;pinsrd xmm1, [eax], 1

				movups xmm2, xmm0
				pxor xmm3, xmm3
				punpcklbw xmm2, xmm3 ;unpackeo para sumar y despues dividir
				movups xmm3, xmm1			
				pxor xmm4, xmm4
				punpcklbw xmm3, xmm4
				paddw xmm2, xmm3 ;tengo la suma de row[col;col+1] y nrow[col;col+1]
				movups xmm3, xmm2
				pshufd xmm3, xmm3, 01001111b
				paddw xmm2, xmm3 ;tengo la suma de los 4 pixels en las posiciones altas
				pxor xmm3, xmm3
				punpckhwd xmm2, xmm3 ;tengo en xmm2 en cada dw r,g,b,a
				cvtdq2ps xmm2, xmm2 ;tengo en xmm2 los valores en flotante
				mov eax, 4
				movd xmm3, eax
				pshufd xmm3, xmm3, 00000000b
				cvtdq2ps xmm3, xmm3 ;tengo los 4 en flotante
				divps xmm2, xmm3; termine la cuenta (divido por 4 cada canal)
				cvtps2dq xmm2, xmm2 ;lo paso a entero de nuevo
			
				;pack AMD
				pxor xmm3, xmm3
				pcmpgtd xmm3, xmm2 ;0 > xmm2?
				mov eax, -1
				movd xmm4, eax
				pshufd xmm4, xmm4, 00000000b	
				pxor xmm3, xmm4			
				pand xmm2, xmm3 ;dejo el valor de los canales que son > 0
				mov eax, 11111111b
				movd xmm3, eax
				pshufd xmm3, xmm3, 00000000b
				pand xmm2, xmm3 ;paso cada canal a byte
				pslld xmm2, 24	
				movups xmm3, xmm2 ;copio
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm2, xmm2, 11111111b
				;packusdw xmm2, xmm2 ;pack a word
				;packuswb xmm2, xmm2 ;pack a byte
				
				;termine un pixel
				mov eax, edi
				add eax, eax
				inc eax
				mov ebx, 4	
				mul ebx		
				add eax, [outrow]
				movd [eax], xmm2 ;guardo el pixel
				inc edi ;icol++
				add DWORD [row], 4
				add DWORD [nrow], 4
				dec ecx
				cmp ecx, 0 			
				jnz .traerDeADos
				jmp .seguir

		.ultimaFilaMenosDe5:
			mov eax, [cols]
			sub eax, edi
			cmp eax, 1
			jz .ultimaFilaCol
		
			dec eax
			mov ecx, eax
			.traerDeADosCols:
				pxor xmm0, xmm0
				pxor xmm1, xmm1
				mov eax, [row]
				;pinsrd AMD				
				movd xmm1, [eax+4]
				por xmm0, xmm1
				pxor xmm1, xmm1				
				movd xmm1, [eax]
				pshufd xmm0, xmm0, 11110011b
				por xmm0, xmm1
				;pinsrb xmm0, [eax], 0
				;inc eax
				;pinsrd xmm0, [eax], 1

				movups xmm2, xmm0
				pxor xmm3, xmm3
				punpcklbw xmm2, xmm3 ;unpackeo para sumar y despues dividir
				movups xmm3, xmm2
				pshufd xmm3, xmm3, 01001111b
				paddw xmm2, xmm3 ;tengo la suma de los 4 pixels en las posiciones altas
				pxor xmm3, xmm3
				punpckhwd xmm2, xmm3 ;tengo en xmm2 en cada dw r,g,b,a
				cvtdq2ps xmm2, xmm2 ;tengo en xmm2 los valores en flotante
				mov eax, 2
				movd xmm3, eax
				pshufd xmm3, xmm3, 00000000b
				cvtdq2ps xmm3, xmm3 ;tengo los 4 en flotante
				divps xmm2, xmm3; termine la cuenta (divido por 4 cada canal)
				cvtps2dq xmm2, xmm2 ;lo paso a entero de nuevo
		
				;pack AMD
				pxor xmm3, xmm3
				pcmpgtd xmm3, xmm2 ;0 > xmm2?
				mov eax, -1
				movd xmm4, eax
				pshufd xmm4, xmm4, 00000000b	
				pxor xmm3, xmm4			
				pand xmm2, xmm3 ;dejo el valor de los canales que son > 0
				mov eax, 11111111b
				movd xmm3, eax
				pshufd xmm3, xmm3, 00000000b
				pand xmm2, xmm3 ;paso cada canal a byte
				pslld xmm2, 24	
				movups xmm3, xmm2 ;copio
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm3, xmm3, 10010011b
				psrld xmm3, 8
				por xmm2, xmm3
				pshufd xmm2, xmm2, 11111111b
				;packusdw xmm2, xmm2 ;pack a word
				;packuswb xmm2, xmm2 ;pack a byte
			
				;termine un pixel
				mov eax, edi
				add eax, eax
				inc eax
				mov ebx, 4	
				mul ebx		
				add eax, [outrow]
				movd [eax], xmm2 ;guardo el pixel
				inc edi ;icol++
				add DWORD [row], 4
				dec ecx
				cmp ecx, 0 			
				jnz .traerDeADosCols

		.ultimaFilaCol:
				;termine un pixel
				mov ecx, [row]
				mov ecx, [ecx]
				mov eax, edi
				add eax, eax
				inc eax
				mov ebx, 4	
				mul ebx		
				add eax, [outrow]
				mov [eax], ecx ;guardo el pixel
				inc edi ;icol++
				add DWORD [row], 4
				jmp .seguir


		.ultimaCol:
			pxor xmm0, xmm0
			pxor xmm1, xmm1
			mov eax, [row]
			;pinsrd AMD				
			movd xmm1, [eax]
			por xmm0, xmm1
			pxor xmm1, xmm1	
			mov eax, [nrow]			
			movd xmm1, [eax]
			pshufd xmm0, xmm0, 11110011b
			por xmm0, xmm1
			;pinsrb xmm0, [eax], 0
			;inc eax
			;pinsrd xmm0, [eax], 1
			
			movups xmm2, xmm0
			pxor xmm3, xmm3
			punpcklbw xmm2, xmm3 ;unpackeo para sumar y despues dividir
			movups xmm3, xmm2
			pshufd xmm3, xmm3, 01001111b
			paddw xmm2, xmm3 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm3, xmm3
			punpckhwd xmm2, xmm3 ;tengo en xmm2 en cada dw r,g,b,a
			cvtdq2ps xmm2, xmm2 ;tengo en xmm2 los valores en flotante
			mov eax, 2
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			cvtdq2ps xmm3, xmm3 ;tengo los 4 en flotante
			divps xmm2, xmm3; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm2, xmm2 ;lo paso a entero de nuevo
			
			;pack AMD
			pxor xmm3, xmm3
			pcmpgtd xmm3, xmm2 ;0 > xmm2?
			mov eax, -1
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b	
			pxor xmm3, xmm4			
			pand xmm2, xmm3 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			pand xmm2, xmm3 ;paso cada canal a byte
			pslld xmm2, 24	
			movups xmm3, xmm2 ;copio
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm2, xmm2, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			;termine un pixel
			mov eax, edi
			add eax, eax
			inc eax
			mov ebx, 4	
			mul ebx		
			add eax, [outrow]
			movd [eax], xmm2 ;guardo el pixel
			inc edi ;icol++
			add DWORD [row], 4
			add DWORD [nrow], 4

			jmp .seguir
		.ultimaFila:
			mov eax, [cols]
			sub eax, edi
			cmp eax, 4
			jz .ultimaFilaMenosDe5
			
			mov eax, [row]
			movups xmm0, [eax]
			pshufd xmm0, xmm0, 00011011b

			;obtener el pixel			
			%rep 3
			movups xmm2, xmm0
			pxor xmm3, xmm3
			punpckhbw xmm2, xmm3 ;unpackeo para sumar y despues dividir
			movups xmm3, xmm2
			pshufd xmm3, xmm3, 01001111b
			paddw xmm2, xmm3 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm3, xmm3
			punpckhwd xmm2, xmm3 ;tengo en xmm2 en cada dw r,g,b,a
			cvtdq2ps xmm2, xmm2 ;tengo en xmm2 los valores en flotante
			mov eax, 2
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			cvtdq2ps xmm3, xmm3 ;tengo los 4 en flotante
			divps xmm2, xmm3; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm2, xmm2 ;lo paso a entero de nuevo
			
			;pack AMD
			pxor xmm3, xmm3
			pcmpgtd xmm3, xmm2 ;0 > xmm2?
			mov eax, -1
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b	
			pxor xmm3, xmm4			
			pand xmm2, xmm3 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm3, eax
			pshufd xmm3, xmm3, 00000000b
			pand xmm2, xmm3 ;paso cada canal a byte
			pslld xmm2, 24	
			movups xmm3, xmm2 ;copio
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm3, xmm3, 10010011b
			psrld xmm3, 8
			por xmm2, xmm3
			pshufd xmm2, xmm2, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			;termine un pixel
			mov eax, edi
			add eax, eax
			inc eax
			mov ebx, 4	
			mul ebx		
			add eax, [outrow]
			movd [eax], xmm2 ;guardo el pixel
			inc edi ;icol++
			pshufd xmm0, xmm0, 10010011b
			%endrep
			add DWORD [row], 12
			add DWORD [nrow], 12
			jmp .seguir
		.seguir:
			jmp .cicloCols

		.termineCols:
			inc esi ;irow++
			jmp .cicloFilas
	.termineFilas:
		pop DWORD [outrow]
		pop DWORD [nrow]
		pop DWORD [row]
		jmp segundaPasada

segundosPix:
	mov eax, [filas]
	add eax, eax
	dec eax
	mov [filas], eax ;filas = 2*filas

	
	mov eax, [cols]
	add eax, eax
	mov [cols], eax ;cols = 2*cols

	xor esi, esi ;irow = 0
	.cicloFilas:
		cmp esi, [filas]
		jge .termineFilas

		mov eax, [widthStepOut]
		mul esi
		add eax, [dataOut]
		mov [prow], eax ;out->imageData + out->widthStep * irow
		mov [outrow], eax

		add eax, [widthStepOut]
		mov [row], eax ;(int)row + out->widthStep
		
		add eax, [widthStepOut]		
		mov [nrow], eax ;(int)row - out->widthStep)

		xor edi, edi
		.cicloCols:
			cmp edi, [cols]
			jge .termineCols ;icol < out->width

			mov eax, [cols]
			sub eax, edi
			cmp eax, 3
			jl .menosDe4Cols ;icol < out->width-3

			mov eax, [prow]
			movups xmm0, [eax]
			mov eax, [row]
			movups xmm1, [eax]

			mov eax, [filas]
			sub eax, esi
			cmp eax, 2
			jl .calculoPixeles ;icol < out->height-2
			mov eax, [nrow]
			movups xmm2, [eax]
			
			.calculoPixeles:
			;calculo a1
			mov eax, -1				
			movd xmm3, eax
			pshufd xmm3, xmm3, 11000000b
			pand xmm3, xmm0
			movd xmm4, eax
			pshufd xmm4, xmm4, 11110011b
			pand xmm4, xmm1
			pshufd xmm4, xmm4, 01111111b
			por xmm3, xmm4	;tengo en xmm3 los 4 pix

			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 4
			cmp esi, 0
			jne .divisionA1
			dec eax
			.divisionA1:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, -1
			movd xmm3, ebx
			pshufd xmm3, xmm3, 00001100b
			pand xmm3, xmm0
			movd xmm4, eax
			pshufd xmm4, xmm4, 11110011b
			por xmm3, xmm4
			movups xmm0, xmm3 ;tengo listo a1

			;calculo a2
			mov eax, -1				
			movd xmm3, eax
			pshufd xmm3, xmm3, 00001111b
			pand xmm3, xmm0
			movd xmm4, eax
			pshufd xmm4, xmm4, 00111111b
			pand xmm4, xmm1
			pshufd xmm4, xmm4, 00001100b
			mov eax, edi
			add eax, 4
			cmp eax, [cols]
			jge .noHayMasColsA2
			mov eax, [prow]
			add eax, 16
			movd xmm5, [eax]
			por xmm4, xmm5
			.noHayMasColsA2:
			por xmm3, xmm4	;tengo en xmm3 los 4 pix
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a			
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 4
			cmp esi, 0
			jne .divisionA2
			dec eax
			.divisionA2:
			mov ebx, edi
			add ebx, 4
			cmp ebx, [cols]
			jl .divisionA22
			dec eax
			.divisionA22:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo		
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, -1
			movd xmm3, ebx
			pshufd xmm3, xmm3, 11000000b
			pand xmm3, xmm0
			movd xmm4, eax
			pshufd xmm4, xmm4, 00111111b
			por xmm3, xmm4
			movups xmm0, xmm3 ;tengo listo a1 y a2

			mov eax, [prow]
			movups [eax], xmm0 ;grabo la fila

			;calculo a3
			mov eax, -1			
			movd xmm3, eax
			pshufd xmm3, xmm3, 11110000b
			pand xmm3, xmm1
			movd xmm4, eax
			pand xmm4, xmm0
			pshufd xmm4, xmm4, 11001111b
			por xmm3, xmm4		
			mov eax, [filas]
			sub eax, esi
			cmp eax, 2
			jl .noHayMasFilasA3 ;icol < out->height-2
			mov eax, -1
			movd xmm4, eax
			pand xmm4, xmm2
			pshufd xmm4, xmm4, 00111111b
			por xmm3, xmm4		;tengo en xmm3 los 4 pix
			.noHayMasFilasA3:
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 4
			cmp edi, 0
			jne .divisionA3
			dec eax ;eax = 3
			.divisionA3:
			mov ebx, [filas]
			sub ebx, esi
			cmp ebx, 2
			jge .divisionA33
			dec eax ;eax = 2
			.divisionA33:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, -1
			movd xmm3, ebx
			pshufd xmm3, xmm3, 00000011b
			pand xmm3, xmm1
			movd xmm4, eax
			por xmm3, xmm4
			movups xmm1, xmm3 ;tengo listo a3

			;calculo a4
			mov eax, -1				
			movd xmm3, eax
			pshufd xmm3, xmm3, 00110011b
			pand xmm3, xmm1
			movd xmm4, eax
			pshufd xmm4, xmm4, 11001111b
			pand xmm4, xmm0
			por xmm3, xmm4		
			mov eax, [filas]
			sub eax, esi
			cmp eax, 2
			jl .noHayMasFilasA4 ;icol < out->height-2
			mov eax, -1
			movd xmm4, eax
			pshufd xmm4, xmm4, 11001111b
			pand xmm4, xmm2
			pshufd xmm4, xmm4, 11111110b
			por xmm3, xmm4		;tengo en xmm3 los 4 pix
			.noHayMasFilasA4:
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 4
			mov ebx, [filas]
			sub ebx, esi
			cmp ebx, 2
			jge .divisionA4
			dec eax ;eax = 3
			.divisionA4:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, -1
			movd xmm3, ebx
			pshufd xmm3, xmm3, 00110000b
			pand xmm3, xmm1
			movd xmm4, eax
			pshufd xmm4, xmm4, 11001111b
			por xmm3, xmm4
			movups xmm1, xmm3 ;tengo listo a4
			
			mov eax, [row]
			movups [eax], xmm1
			jmp .seguir

		.menosDe4Cols:
			;calculo M1
			pxor xmm3, xmm3
			mov eax, [prow]				
			movd xmm4, [eax]
			por xmm3, xmm4
			pshufd xmm3, xmm3, 10010011b
			add eax, 4
			movd xmm4, [eax]			
			por xmm3, xmm4
			pshufd xmm3, xmm3, 10010011b
			mov eax, [row]
			add eax, 4
			movd xmm4, [eax]
			por xmm3, xmm4

			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 3
			cmp edi, 0
			jg .divisionM4
			dec eax ;eax = 3
			.divisionM4:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte

			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, [prow]
			add ebx, 4
			mov [ebx], eax

			;calculo M2
			pxor xmm3, xmm3
			mov eax, [prow]				
			movd xmm4, [eax]
			por xmm3, xmm4
			pshufd xmm3, xmm3, 10010011b
			mov eax, [row]
			movd xmm4, [eax]			
			por xmm3, xmm4
			pshufd xmm3, xmm3, 10010011b
			add eax, 4
			movd xmm4, [eax]
			por xmm3, xmm4			
			pshufd xmm3, xmm3, 10010011b
			mov eax, [filas]
			sub eax, esi
			cmp eax, 2
			jl .noHayMasFilasM2 ;icol < out->height-2
			mov eax, [nrow]
			movd xmm4, [eax]
			por xmm3, xmm4
			.noHayMasFilasM2:
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			pxor xmm5, xmm5
			punpckhbw xmm4, xmm5 ;tengo en xmm4 unpackeado para sumar los pix 0 y 1
			punpckhbw xmm3, xmm5
			paddw xmm3, xmm4 ;tengo la suma de pix 0 al 3
			movups xmm4, xmm3
			pshufd xmm4, xmm4, 01001111b
			paddw xmm3, xmm4 ;tengo la suma de los 4 pixels en las posiciones altas
			pxor xmm4, xmm4
			punpckhwd xmm3, xmm4 ;tengo en xmm3 en cada dw r,g,b,a
			cvtdq2ps xmm3, xmm3 ;tengo en xmm3 los valores en flotante
			mov eax, 4
			cmp edi, 0
			jne .divisionM2
			dec eax ;eax = 3
			.divisionM2:
			mov ebx, [filas]
			sub ebx, esi
			cmp ebx, 2
			jge .divisionM22
			dec eax ;eax = 2
			.divisionM22:
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			cvtdq2ps xmm4, xmm4 ;tengo los 4 en flotante
			divps xmm3, xmm4; termine la cuenta (divido por 4 cada canal)
			cvtps2dq xmm3, xmm3 ;lo paso a entero de nuevo
			;pack AMD
			pxor xmm4, xmm4
			pcmpgtd xmm4, xmm3 ;0 > xmm2?
			mov eax, -1
			movd xmm5, eax
			pshufd xmm5, xmm5, 00000000b	
			pxor xmm4, xmm5			
			pand xmm3, xmm4 ;dejo el valor de los canales que son > 0
			mov eax, 11111111b
			movd xmm4, eax
			pshufd xmm4, xmm4, 00000000b
			pand xmm3, xmm4 ;paso cada canal a byte
			pslld xmm3, 24	
			movups xmm4, xmm3 ;copio
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm4, xmm4, 10010011b
			psrld xmm4, 8
			por xmm3, xmm4
			pshufd xmm3, xmm3, 11111111b
			;packusdw xmm2, xmm2 ;pack a word
			;packuswb xmm2, xmm2 ;pack a byte
			
			movd eax, xmm3 ;tengo el pix a copiar
			mov ebx, [row]
			mov [ebx], eax

			mov eax, esi
			add eax, 2
			cmp eax, [filas]
			jge .noCopiarAbajoM
			mov eax, [nrow]
			add eax, 4
			mov ebx, [row]
			add ebx, 4
			mov ebx, [ebx]
			mov [eax], ebx

			.noCopiarAbajoM:
				jmp .noCopiarAbajo
		.seguir:
			mov eax, edi
			add eax, 4
			cmp eax, [cols]
			jge .noCopiarALaDerecha 
			movups xmm3, xmm1
			pshufd xmm3, xmm3, 11100111b
			movd ebx, xmm3
			mov eax, [row]
			add eax, 16
			mov [eax], ebx
		.noCopiarALaDerecha:
			mov eax, esi
			add eax, 2
			cmp eax, [filas]
			jge .noCopiarAbajo
			pshufd xmm1, xmm1, 11100101b			
			movd ebx, xmm1
			mov eax, [nrow]
			add eax, 4
			mov [eax], ebx
			pshufd xmm1, xmm1, 11100111b
			movd ebx, xmm1
			add eax, 8
			mov [eax], ebx
		.noCopiarAbajo:
			add DWORD [prow], 16
			add DWORD [row], 16
			add DWORD [nrow], 16
			add edi, 4
			jmp .cicloCols

		.termineCols:
			add esi, 2 ;irow+= 2
			jmp .cicloFilas
	.termineFilas:
		jmp termine
