__device__
bool obtener_de_compartida(unsigned int fila, unsigned int col)
{
	unsigned int colMin, colMax, filaMin, filaMax;
	colMin = blockIdx.x * blockDim.x;
	colMax = (blockIdx.x + 1) * blockDim.x;
	filaMin = blockIdx.y * blockDim.y;
	filaMax = (blockIdx.y + 1) * blockDim.y;
	
	return (colMin <= col && col < colMax && filaMin <= fila && fila < filaMax);
}

__global__
void zoomInterpolador_matriz_fase_2( unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStepIn, unsigned int height, unsigned int widthStepOut )
{
	unsigned int col = (blockIdx.x * blockDim.x + threadIdx.x) * 4;
	unsigned int fila = blockIdx.y * blockDim.y + threadIdx.y;
	unsigned int indexOut;

	if (fila < height && col < widthStepIn)
	{
		//Fase 2: Proceso los pixeles que faltan.
		indexOut =  2 * col + 2 * fila * widthStepOut;
		unsigned int indexOutPar = indexOut + 4;
		unsigned int indexOutImpar = indexOut + widthStepOut;

		if (col == 0 && fila < height - 1)
		{
			for(int i = 0; i < 4; i++)
			{
				if (fila == 0)
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4] + imgOut[indexOutPar+i+4])/3;
				else
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4] + imgOut[indexOutPar+i+4])/4;
				imgOut[indexOutImpar + i] = (imgOut[indexOutImpar - widthStepOut + i] + imgOut[indexOutImpar + widthStepOut + i] + imgOut[indexOutImpar + i + 4])/3;
			}
		}
		else if (col < widthStepIn - 4)
		{
			if (fila == 0)
			{
				for(int i = 0; i < 4; i++)
				{
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4] + imgOut[indexOutPar+i+4])/3;
					imgOut[indexOutImpar+i] = (imgOut[indexOutImpar+i - widthStepOut] + imgOut[indexOutImpar+i + widthStepOut] + imgOut[indexOutImpar+i-4] + imgOut[indexOutImpar+i+4])/4;
				}
			}
			else if (fila < height -1)
			{
				for(int i = 0; i < 4; i++)
				{
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i - widthStepOut] + imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4] + imgOut[indexOutPar+i+4])/4;

					imgOut[indexOutImpar+i] = (imgOut[indexOutImpar+i - widthStepOut] + imgOut[indexOutImpar+i + widthStepOut] + imgOut[indexOutImpar+i-4] + imgOut[indexOutImpar+i+4])/4;
				}
			}
			else
			{
				for(int i = 0; i < 4; i++)
				{
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i - widthStepOut] + imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4] + imgOut[indexOutPar+i+4])/4;
					imgOut[indexOutImpar+i] = (imgOut[indexOutImpar+i - widthStepOut] + imgOut[indexOutImpar+i-4] + imgOut[indexOutImpar+i+4])/3;
				}
			}
		}
		else
		{
			if (fila == 0)
			{
				for (int i = 0; i < 4; i++)
				{
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4])/2;
					imgOut[indexOutImpar+i] = (imgOut[indexOutImpar+i - widthStepOut] + imgOut[indexOutImpar+i + widthStepOut] + imgOut[indexOutImpar+i-4])/3;
				}
			}
			else if (fila < height)
			{
				for (int i = 0; i < 4; i++)
				{
					imgOut[indexOutPar+i] = (imgOut[indexOutPar+i - widthStepOut] + imgOut[indexOutPar+i + widthStepOut] + imgOut[indexOutPar+i-4])/3;
					imgOut[indexOutImpar+i] = (imgOut[indexOutImpar+i - widthStepOut] + imgOut[indexOutImpar+i + widthStepOut] + imgOut[indexOutImpar+i-4])/3;
				}
			}
		}
	}
}

__global__
void zoomInterpolador_matriz_fase_1( unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStepIn, unsigned int height, unsigned int widthStepOut )
{
	//Memoria compartida por el bloque de threads.
	__shared__ unsigned char memBloque[16 * 4 * 16];

	unsigned int col = (blockIdx.x * blockDim.x + threadIdx.x) * 4;
	unsigned int fila = blockIdx.y * blockDim.y + threadIdx.y;
	unsigned int indexIn = col + fila * widthStepIn;
	unsigned int indexOut;
	unsigned int indexBloque = blockDim.x * threadIdx.y * 4 + threadIdx.x * 4;

	//Paso de la memoria global de la placa a la memoria del bloque.
	if (fila < height && col < widthStepIn)
	{
		for(int i = 0; i < 4; i++)
			memBloque[indexBloque+i] = imgIn[indexIn+i];
	}
	//Espero a que lleguen todos los threads del bloque.
	__syncthreads();

	if (fila < height && col < widthStepIn)
	{
		//Copio la imagen original.
		indexOut =  2 * col + 2 * fila * widthStepOut;
		for(int i = 0; i < 4; i++)
			imgOut[indexOut+i] = memBloque[indexBloque+i];

		for(int i = 4; i < 8; i++)
			imgOut[indexOut+i] = 0;

		//Fase 1: Proceso los pixeles que estan rodeados por la imagen.
		indexOut = 2 * col + 4 + 2 * fila * widthStepOut + widthStepOut;
		for(int i = 0; i < 4; i++)
		{
			if (col < widthStepIn - 4)
			{
				if (fila < height - 1)
				{
					imgOut[indexOut+i] = ((obtener_de_compartida(fila, (col / 4) + 1 ) ? memBloque[indexBloque + 4 + i] : imgIn[indexIn + 4 + i]) + (obtener_de_compartida(fila, (col / 4) ) ? memBloque[indexBloque + i] : imgIn[indexIn + i]) + (obtener_de_compartida(fila + 1, (col / 4) + 1 ) ? memBloque[indexBloque + blockDim.x * 4 + 4 + i] : imgIn[indexIn + widthStepIn + 4 + i]) + (obtener_de_compartida(fila + 1, (col / 4)) ? memBloque[indexBloque + blockDim.x * 4 + i] : imgIn[indexIn + widthStepIn + i])) / 4;
				}
				else
				{
					imgOut[indexOut+i] = ((obtener_de_compartida(fila, (col / 4) + 1 ) ? memBloque[indexBloque + 4 + i] : imgIn[indexIn + 4 + i]) + (obtener_de_compartida(fila, (col / 4) ) ? memBloque[indexBloque + i] : imgIn[indexIn + i])) / 2;
				}
			}	
			else
			{
				if (fila < height -1)
				{
					imgOut[indexOut+i] = ((obtener_de_compartida(fila, (col / 4) ) ? memBloque[indexBloque + i] : imgIn[indexIn + i]) + (obtener_de_compartida(fila + 1, (col / 4)) ? memBloque[indexBloque + blockDim.x * 4 + i] : imgIn[indexIn + widthStepIn + i])) / 2;
				}
				else
				{
					imgOut[indexOut+i] = obtener_de_compartida(fila, (col / 4) ) ? memBloque[indexBloque + i] : imgIn[indexIn + i];
				}
			}
		}
	}
}

extern "C" void zoomInterpoladorCUDA(unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStep, unsigned int height, unsigned int widthStepOut );

void zoomInterpoladorCUDA(unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStep, unsigned int height, unsigned int widthStepOut ) 
{
	const int blockWidth = 16;	
	const int blockHeight = 16;

	unsigned char* in;
	unsigned char* out;
	const unsigned int size = widthStep * height * sizeof(unsigned char);
	
	cudaMalloc( (void**)&in, size );
	cudaMalloc( (void**)&out, widthStepOut * height * 2 * sizeof(unsigned char) );
	cudaMemcpy( in, imgIn, size, cudaMemcpyHostToDevice );

	dim3 dimBlock( blockWidth, blockHeight );
	dim3 dimGrid( (widthStep / (4 * dimBlock.x))+1, (height / dimBlock.y) + 1 );

	zoomInterpolador_matriz_fase_1<<<dimGrid, dimBlock>>>( in, out, widthStep, height, widthStepOut );
	zoomInterpolador_matriz_fase_2<<<dimGrid, dimBlock>>>( in, out, widthStep, height, widthStepOut );

	cudaMemcpy( imgOut, out, widthStepOut * height * 2 * sizeof(unsigned char), cudaMemcpyDeviceToHost );
	cudaFree( in ); cudaFree( out );
}
