__global__ void cu_laplace( unsigned char* in, unsigned char* out, unsigned int step, unsigned int height,unsigned int width )
{
	__shared__ unsigned char fast_mem[18 * 18 * 3]; //traigo los pixels correspondientes al bloque, y a los vecinos

	unsigned int col = blockIdx.x * blockDim.x + threadIdx.x;
	unsigned int fila = blockIdx.y * blockDim.y + threadIdx.y;

	unsigned char *p_i = in + fila * step + col * 3 ;
	unsigned char *p_o = out + fila * step + col * 3 ;

	bool en_rango = fila < height && col < width;
	bool soy_limite = fila == 0 || fila == (height-1) || col == 0 || col == (width-1);

	if(en_rango) {
		//Paso de la memoria global de la placa a la memoria del bloque.
		int fast_mem_off = 3 * ((threadIdx.y + 1) * 18 + threadIdx.x + 1);
		for(int i = 0; i < 3; i++) fast_mem[fast_mem_off + i] = p_i[i];

		//cargo los pixels q estan a la izquierda
		if(threadIdx.x == 0 && blockIdx.x != 0) {	
			int fast_mem_off = 3 * ((threadIdx.y + 1) * 18);
			for(int i = 0; i < 3; i++)
				fast_mem[fast_mem_off + i] = (p_i-3)[i];
		}

		//cargo los pixels q estan a la derecha 
		if(threadIdx.x == (blockDim.x-1) && blockIdx.x != (gridDim.x - 1)) {	
			int fast_mem_off = 3 * ((threadIdx.y + 1) * 18 + 16 + 1); //16 + 1 es de los 16 pixels de ancho del bloque, mas la barra de la izquierda
			for(int i = 0; i < 3; i++)
				fast_mem[fast_mem_off + i] = (p_i+3)[i];
		}

		//cargo los pixels q estan arriba
		if(threadIdx.y == 0 && blockIdx.y != 0) {	
			int fast_mem_off = 3 * ( threadIdx.x + 1 );
			for(int i = 0; i < 3; i++)
				fast_mem[fast_mem_off + i] = (p_i-step)[i];
		}

		//cargo los pixels q estan abajo
		if(threadIdx.y == (blockDim.y - 1) && blockIdx.y != (gridDim.y - 1)) {	
			int fast_mem_off = 3 * ((16 + 1) * 18 + threadIdx.x + 1 );
			for(int i = 0; i < 3; i++)
				fast_mem[fast_mem_off + i] = (p_i+step)[i];
		}

		//las esquinas se tratan a parte
		if(threadIdx.x == 0 && threadIdx.y == 0 && blockIdx.x != 0 && blockIdx.y != 0) {
			unsigned char* bad_corner = p_i - step - 3;
			for(int i = 0; i < 3;++i) fast_mem[i] = bad_corner[i];
		} else if(threadIdx.x == (blockDim.x - 1) && threadIdx.y == 0 && blockIdx.x != (gridDim.x - 1) && blockIdx.y != 0) {
			int fast_mem_off = 3 * 17; 
			unsigned char* bad_corner = p_i - step + 3;
			for(int i = 0; i < 3;++i) fast_mem[fast_mem_off + i] = bad_corner[i];
		} else if(threadIdx.x == 0 && threadIdx.y == (blockDim.y - 1) && blockIdx.x != 0 && blockIdx.y != (gridDim.y-1)) {
			int fast_mem_off = 3 * 18 * 17; 
			unsigned char* bad_corner = p_i + step - 3;
			for(int i = 0; i < 3;++i) fast_mem[fast_mem_off + i] = bad_corner[i];
		} else if(threadIdx.x == (blockDim.x - 1) && threadIdx.y == (blockDim.y - 1) && blockIdx.x != (gridDim.x-1) && blockIdx.y != (gridDim.y -1)) {
			int fast_mem_off = 3 * (18 * 17 + 17); 
			unsigned char* bad_corner = p_i + step + 3;
			for(int i = 0; i < 3;++i) fast_mem[fast_mem_off + i] = bad_corner[i];
		}
	}

	//Espero a que lleguen todos los threads del bloque.
	__syncthreads();
	#define __indx(x,y) (3 * (x) + 3 * 18 * (y))

	if(en_rango) {
		if(!soy_limite) {
			int local_in[] = {0, 0, 0};
			int my_fast_mem_off = 3 * ((threadIdx.y + 1) * 18 + threadIdx.x + 1); //tiene el offset de mi posicion en la memoria compartida

			for(int i = 0; i < 3;++i) local_in[i] = 8 * fast_mem[my_fast_mem_off + i + __indx(0,0)];

			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx(-1, 0)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx( 1, 0)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx( 0,-1)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx( 0, 1)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx( 1, 1)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx( 1,-1)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx(-1, 1)];
			for(int i = 0; i < 3;++i) local_in[i] -= fast_mem[my_fast_mem_off + i + __indx(-1,-1)];

			for(int i = 0; i < 3;++i) p_o[i] = min(max(local_in[i],0),255);
		} else {
			for(int i = 0; i < 3;++i) p_o[i] = 0;
		}
	}
}

__global__ void cu_laplacesum(unsigned char *cu_in,unsigned char *cu_out,unsigned int step,unsigned int height,unsigned int width) {
	unsigned int n_pixel = blockIdx.x * blockDim.x + threadIdx.x; //me da en que numero de pixel estoy!
	if(n_pixel < height * width) {
		unsigned int n_row,n_col;
		unsigned char* pix_in,*pix_out;

		n_row = n_pixel / width;
		n_col = n_pixel % width;

		pix_in = cu_in + n_row * step + 3 * sizeof(unsigned char) * n_col;
		pix_out = cu_out + n_row * step + 3 * sizeof(unsigned char) * n_col;

		pix_out[0] = min(max(pix_in[0] + pix_out[0],0),255);
		pix_out[1] = min(max(pix_in[1] + pix_out[1],0),255);
		pix_out[2] = min(max(pix_in[2] + pix_out[2],0),255);
	}
}

unsigned char *cu_mem_in = 0, *cu_mem_out = 0;
bool cu_mem_init = false;

extern "C" void laplace(unsigned char* in_data, unsigned char* out_data, unsigned int step, unsigned int height ,unsigned int width) 
{
	const int blockWidth = 16;	
	const int blockHeight = 16;

	if(!cu_mem_init){
		cudaMalloc((void**)&cu_mem_in,sizeof(unsigned char) * step * height);
		cudaMalloc((void**)&cu_mem_out,sizeof(unsigned char) * step * height);
		cu_mem_init = true;
	}

	dim3 dimBlock( blockWidth, blockHeight );
	dim3 dimGrid( (width / ( dimBlock.x))+1, (height / dimBlock.y) + 1 );

	cudaMemcpy(cu_mem_in,in_data,sizeof(unsigned char) * step * height,cudaMemcpyHostToDevice);

	cu_laplace<<<dimGrid , dimBlock>>>(cu_mem_in,cu_mem_out,step,height,width);

	cudaMemcpy(out_data,cu_mem_out,sizeof(unsigned char) * step * height,cudaMemcpyDeviceToHost);
}

extern "C" void laplacesum(unsigned char* in_data_1,unsigned char* in_data_2,unsigned char* out_data,unsigned int step,unsigned int height,unsigned int width) {
	if(!cu_mem_init){
		cudaMalloc((void**)&cu_mem_in,sizeof(unsigned char) * step * height);
		cudaMalloc((void**)&cu_mem_out,sizeof(unsigned char) * step * height);
		cu_mem_init = true;
	}

	cudaMemcpy(cu_mem_in,in_data_1,sizeof(unsigned char) * step * height,cudaMemcpyHostToDevice);
	cudaMemcpy(cu_mem_out,in_data_2,sizeof(unsigned char) * step * height,cudaMemcpyHostToDevice);

	dim3 dimBlock(16*16,1);
	dim3 dimGrid((width * height)/dimBlock.x + 1,1); //proceso de a tiras.

	cu_laplacesum<<<dimGrid , dimBlock>>>(cu_mem_in,cu_mem_out,step,height,width);

	cudaMemcpy(out_data,cu_mem_out,sizeof(unsigned char) * step * height,cudaMemcpyDeviceToHost);
}

extern "C" void cuda_clean() {
	if(cu_mem_init) {
		cudaFree(cu_mem_in);
		cudaFree(cu_mem_out);
	}
}
