__global__
void pixelizar_matriz( unsigned int* imgIn, unsigned int* imgOut, unsigned int widthStep, unsigned int height )
{
	//Memoria compartida por el bloque de threads.
	__shared__ unsigned int memBloque[16 * 16];

	int col = blockIdx.x * blockDim.x + threadIdx.x;
	int fila = blockIdx.y * blockDim.y + threadIdx.y;
	int index = col + fila * widthStep / 4;
	int indexBloque = blockDim.x * threadIdx.y + threadIdx.x;

	//Paso de la memoria global de la placa a la memoria del bloque.
	if (fila < height && col < (widthStep / 4) && indexBloque % 8 == 0)
		memBloque[indexBloque] = imgIn[index];

	//Espero a que lleguen todos los threads del bloque.
	__syncthreads();

	if (fila < height && col < (widthStep / 4))
	{
		if (threadIdx.x % 8 != 0)
		{
			if (threadIdx.y % 8 != 0)
				imgOut[index] = memBloque[indexBloque - (threadIdx.x % 8) - blockDim.x * (threadIdx.y % 8)];	
			else
				imgOut[index] = memBloque[indexBloque - (threadIdx.x % 8)];
		}
		else
		{
			if (threadIdx.y % 8 == 0)
				imgOut[index] = memBloque[indexBloque];
			else
				imgOut[index] = memBloque[indexBloque - blockDim.x * (threadIdx.y % 8)];	
		}

	}
}

extern "C" void pixelizarCUDA(unsigned int* imgIn, unsigned int* imgOut, unsigned int widthStep, unsigned int height );

void pixelizarCUDA(unsigned int* imgIn, unsigned int* imgOut, unsigned int widthStep, unsigned int height ) 
{
	const int blockWidth = 16;	
	const int blockHeight = 16;

	unsigned int* in;
	unsigned int* out;
	const unsigned int size = widthStep * height * sizeof(unsigned int) / 4;
	
	cudaMalloc( (void**)&in, size );
	cudaMalloc( (void**)&out, size );
	cudaMemcpy( in, imgIn, size, cudaMemcpyHostToDevice );

	dim3 dimBlock( blockWidth, blockHeight );
	dim3 dimGrid( (widthStep / (4 * dimBlock.x))+1, (height / dimBlock.y) + 1 );

	pixelizar_matriz<<<dimGrid, dimBlock>>>( in, out, widthStep, height );

	cudaMemcpy( imgOut, out, size, cudaMemcpyDeviceToHost );
	cudaFree( in ); cudaFree( out );
}
