#include<iostream>
#include"utils.h"

using namespace std;

#ifdef _Asm
extern "C" void rangeOperator(IplImage* in, IplImage* out);
#elif _Cuda 
extern "C" {
	void rangeOperator(unsigned char*,unsigned char*,unsigned int,unsigned int,unsigned int);
	void cuda_clean();
}
#else
void rangeOperator(IplImage* in, IplImage* out)
{
	pixel* dataOut = (pixel*)out->imageData;	
	pixel* dataIn = (pixel*)in->imageData;	
	
	pixel* p_inrow_up; //Puntero a la fila anterior a la que estoy leyendo.
	pixel* p_inrow; //Puntero a la fila que estoy leyendo.
	pixel* p_inrow_down; //Puntero a la fila siguiente a la que estoy leyendo.
	pixel* p_outrow; //Puntero a la fila que estoy escribiendo.
	int filas = in->height; //Cantidad de filas.
	int cols = in->width; //Cantidad de columnas.
	int widthStep = in->widthStep; //Cantidad de bytes por fila.	
	unsigned char maxR, minR, maxB, minB, maxG, minG, temp;
	
	for(int irow = 1;irow < filas-1;irow++){
		p_inrow_up = (pixel*)((int)dataIn + widthStep * (irow-1)); //calculo cual es la proxima columna
		p_inrow = (pixel*)((int)dataIn + widthStep * irow); //calculo cual es la proxima columna
		p_inrow_down = (pixel*)((int)dataIn + widthStep * (irow+1)); //calculo cual es la proxima columna
		p_outrow = (pixel*)((int)dataOut + widthStep * irow); 

		for(int icol = 1;icol < cols-1;icol++){
			maxG = maxR = maxB = 0;
			minB = minR = minG = 255;
			for(int x = icol-1; x<icol+2; x++){
				temp = max(p_inrow_up[x].r, p_inrow[x].r);
				temp = max(temp, p_inrow_down[x].r);
				maxR = max(temp, maxR);
				temp = min(p_inrow_up[x].r, p_inrow[x].r);
				temp = min(temp, p_inrow_down[x].r);
				minR = min(temp, minR);
				temp = max(p_inrow_up[x].b, p_inrow[x].b);
				temp = max(temp, p_inrow_down[x].b);
				maxB = max(temp, maxB);
				temp = min(p_inrow_up[x].b, p_inrow[x].b);
				temp = min(temp, p_inrow_down[x].b);
				minB = min(temp, minB);
				temp = max(p_inrow_up[x].g, p_inrow[x].g);
				temp = max(temp, p_inrow_down[x].g);
				maxG = max(temp, maxG);
				temp = min(p_inrow_up[x].g, p_inrow[x].g);
				temp = min(temp, p_inrow_down[x].g);
				minG = min(temp, minG);
			}
			p_outrow[icol].b = maxB-minB; 
			p_outrow[icol].g = maxG-minG; 
			p_outrow[icol].r = maxR-minR;
		}
	}
	
}
#endif

#ifdef _Video
int main( int argc, char** argv )
{
	bool isinit = false;
	IplImage  *frame;
	IplImage  *frameAlpha;
	IplImage  *salida;
	int       key;
   
	//Chequeo que venga la ruta del AVI
	assert( argc == 2 );
   
	//Cargo el AVI
	CvCapture *capture = cvCaptureFromAVI( argv[1] );
   
	//Si no lo pude cargar salgo	
	if( !capture ){
		cout <<  "No se pudo abrir el archivo " << argv[1] << endl;
		return -1;
	}    
   
	//Reviso la cantidad de fps que tengo que mostrar
	int fps = (int)cvGetCaptureProperty( capture, CV_CAP_PROP_FPS );
   
	//Creo la ventana para mostrarlo
	cvNamedWindow( "Video Original", CV_WINDOW_AUTOSIZE );
	cvNamedWindow( "Video Salida", CV_WINDOW_AUTOSIZE );

   
	//Mientras no toquen 'q'
	while(1){
        	//Traigo un frame
		frame = cvQueryFrame( capture );
            
        	//Si no pude traer nada, se acabó el video
	        if( !frame ){
	         cvSetCaptureProperty(capture, CV_CAP_PROP_POS_AVI_RATIO, 0);
	         frame = cvQueryFrame( capture );
	        }
	        if( !frame ) break;
		
		if(!isinit){
#ifdef _Asm	        
			frameAlpha = agregarCanalAlpha(frame);
			salida = chequearImagen(frameAlpha);
#else
			salida = chequearImagen(frame);
#endif
			isinit = true;
		}

#ifdef _Asm	        
		reusarCanalAlfa(frame,frameAlpha);
		rangeOperator(frameAlpha, salida);
#endif

#ifdef _Cpp
		rangeOperator(frame, salida);
#endif

#ifdef _Cuda
	rangeOperator((unsigned char*)frame->imageData,(unsigned char*)salida->imageData,frame->widthStep,frame->height,frame->width);
#endif
		//Muestro el frame
		cvShowImage( "Video Original", frame );
		cvShowImage( "Video Salida", salida );

        	//Espero el tiempo que calcule antes para mostrar el próximo frame
		key = cvWaitKey( 1000 / fps ) & 0xFF;
		if(key == 27) break;
	}
   
	//Limpio!
	cvReleaseCapture(&capture);
#ifdef _Asm
	cvReleaseImage(&frameAlpha);
#endif
	cvReleaseImage(&salida);
	limpiarVentanas();
#ifdef _Cuda
	cuda_clean();
#endif
	return 0;
}


#else


int main(int argc,char** argv){
	if(argc<2)return 0;
	if(!strcmp(argv[1],"--help")){
		cout << "./rangeOperator [imagein]" << endl; 
		cout << "./rangeOperator [imagein] [imageout]" << endl; 
		return 0;
	}

	const char* imagename = argv[1]; 
	const char* outname = (argc<3)?0:argv[2];
	
	IplImage* img = cvLoadImage(imagename);
	IplImage* res = 0;

	ERROR(img,"No se pudo abrir la imagen");
#ifdef _Asm
	{
		IplImage* temp= 0;
		temp = agregarCanalAlpha(img);
		cvReleaseImage(&img);
		img = temp;
	}
#endif
	res = chequearImagen(img);

#ifdef _Cuda
	rangeOperator((unsigned char*)img->imageData,(unsigned char*)res->imageData,img->widthStep,img->height,img->width);
#else
	rangeOperator(img,res);
#endif

	if(!outname){
		mostrarVentanas(img,res);
		limpiarVentanas();
	}else{
		cvSaveImage(outname,res);
		cout << "Imagen guardada como " << outname << endl;
	}

#ifdef _Cuda
	cuda_clean();
#endif
	limpiarImagenes(img,res);
	return 0;
}
#endif
