#define pixel_size 3
#define techo(expr) (min(255.0,max(0.0,(expr))))

__global__ void cu_rgbtohsl(unsigned char* cu_in,unsigned char* cu_out,unsigned int step,unsigned int height,unsigned int width,int sat) {
	unsigned int n_pixel = blockIdx.x * blockDim.x + threadIdx.x; //me da en que numero de pixel estoy!
	if(n_pixel < height * width) {
		unsigned int n_row,n_col;
		unsigned char* pix_in,*pix_out;
		unsigned char local_pix[3];

		n_row = n_pixel / width;
		n_col = n_pixel % width;

		pix_in = cu_in + n_row * step + pixel_size * sizeof(unsigned char) * n_col;
		pix_out = cu_out + n_row * step + pixel_size * sizeof(unsigned char) * n_col;
		
		local_pix[0] = pix_in[0];
		local_pix[1] = pix_in[1];
		local_pix[2] = pix_in[2];
		
		float r, g, b, h, s, l; 
		r = local_pix[0] / 256.0; 
              	g = local_pix[1] / 256.0; 
              	b = local_pix[2] / 256.0;
                          
              	float maxColor = max(r, max(g, b)); 
              	float minColor = min(r, min(g, b));
                          
                if(minColor == maxColor){
        		h = 0.0; 
		        s = 0.0;       
        	      	l = r; 
                }else{   
                        l = (minColor + maxColor) / 2;     
          
                        if(l < 0.5){
                        	s = (maxColor - minColor) / (maxColor + minColor);
                        }else{
	                        s = (maxColor - minColor) / (2.0 - maxColor - minColor);
                        }

			if(r == maxColor){
				h = (g - b) / (maxColor - minColor);
			}else if(g == maxColor){
				h = 2.0 + (b - r) / (maxColor - minColor);
			}else{
				h = 4.0 + (r - g) / (maxColor - minColor);
			}
			h /= 6;
                                  
			if(h < 0) h ++;
		}
                          
               	local_pix[0] = int(h * 255.0);
               	local_pix[1] = int(l * 255.0);
               	local_pix[2] = int( techo( s * 255.0 + sat));

		__syncthreads(); //como pueden des-sincronizarse los threads por el branching anterior, esperan.

		float temp1, temp2, tempr, tempg, tempb;
		h = local_pix[0] / 256.0;
		l = local_pix[1] / 256.0;
		s = local_pix[2] / 256.0;

		if(s == 0){
			r = g = b = l;
		}else{
			if(l < 0.5) temp2 = l * (1 + s);      
			else temp2 = (l + s) - (l * s);     
			temp1 = 2 * l - temp2;    
			tempr = h + 1.0 / 3.0;    
			if(tempr > 1) tempr--;
			tempg = h;     
			tempb = h - 1.0 / 3.0;
			if(tempb < 0) tempb++; 
        
			if(tempr < 1.0 / 6.0) r = temp1 + (temp2 - temp1) * 6.0 * tempr;      
			else if(tempr < 0.5) r = temp2;   
			else if(tempr < 2.0 / 3.0) r = temp1 + (temp2 - temp1) * ((2.0 / 3.0) - tempr) * 6.0;
			else r = temp1; 
        
			if(tempg < 1.0 / 6.0) g = temp1 + (temp2 - temp1) * 6.0 * tempg;    
			else if(tempg < 0.5) g = temp2;
			else if(tempg < 2.0 / 3.0) g = temp1 + (temp2 - temp1) * ((2.0 / 3.0) - tempg) * 6.0;
			else g = temp1; 
        
			if(tempb < 1.0 / 6.0) b = temp1 + (temp2 - temp1) * 6.0 * tempb;   
			else if(tempb < 0.5) b = temp2; 
			else if(tempb < 2.0 / 3.0) b = temp1 + (temp2 - temp1) * ((2.0 / 3.0) - tempb) * 6.0;    
			else b = temp1;
		}

		pix_out[0] = int(r * 255.0);
		pix_out[1] = int(g * 255.0);
		pix_out[2] = int(b * 255.0);
	}
}

unsigned char *cu_mem_in = 0, *cu_mem_out = 0;
bool cu_mem_init = false;

extern "C" void rgbtohsl(unsigned char* in_data,unsigned char* out_data,unsigned int width,unsigned int height,unsigned int step,int sat) {
	if(!cu_mem_init){
		cudaMalloc((void**)&cu_mem_in,sizeof(unsigned char) * step * height);
		cudaMalloc((void**)&cu_mem_out,sizeof(unsigned char) * step * height);
		cu_mem_init = true;
	}

	cudaMemcpy(cu_mem_in,in_data,sizeof(unsigned char) * step * height,cudaMemcpyHostToDevice);
	dim3 dimBlock(16*16,1);
	dim3 dimGrid((width * height)/dimBlock.x + 1,1); //proceso de a tiras.

	cu_rgbtohsl<<<dimGrid , dimBlock>>>(cu_mem_in,cu_mem_out,step,height,width,sat);

	cudaMemcpy(out_data,cu_mem_out,sizeof(unsigned char) * step * height,cudaMemcpyDeviceToHost);
}

extern "C" void cuda_clean() {
	if(cu_mem_init) {
		cudaFree(cu_mem_in);
		cudaFree(cu_mem_out);
	}
}
