__device__
bool obtener_de_compartida(unsigned int fila, unsigned int col)
{
	unsigned int colMin, colMax, filaMin, filaMax;
	colMin = blockIdx.x * blockDim.x;
	colMax = (blockIdx.x + 1) * blockDim.x;
	filaMin = blockIdx.y * blockDim.y;
	filaMax = (blockIdx.y + 1) * blockDim.y;
	
	return (colMin <= col && col < colMax && filaMin <= fila && fila < filaMax);
}

__global__
void sobel_matriz( unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStep, unsigned int height )
{
	//Memoria compartida por el bloque de threads.
	__shared__ unsigned char memBloque[16 * 4 * 16];

	unsigned int col = (blockIdx.x * blockDim.x + threadIdx.x) * 4;
	unsigned int fila = blockIdx.y * blockDim.y + threadIdx.y;
	unsigned int index = col + fila * widthStep;
	unsigned int indexBloque = blockDim.x * threadIdx.y * 4 + threadIdx.x * 4;

	//Paso de la memoria global de la placa a la memoria del bloque.
	if (fila < height && col < widthStep)
	{
		for(int i = 0; i < 4; i++)
			memBloque[indexBloque+i] = imgIn[index+i];
	}
	//Espero a que lleguen todos los threads del bloque.
	__syncthreads();

	int resultado;
	unsigned char kernel0 [3];
	unsigned char kernel1 [3];

	if (fila < height - 1 && col < widthStep - 4 && fila != 0 && col != 0)
	{
		for(int i = 0; i < 3; i++)
		{
			resultado = (obtener_de_compartida(fila - 1, (col / 4) - 1 ) ? memBloque[indexBloque - blockDim.x * 4 - 4 + i] : imgIn[index - widthStep - 4 + i]) + 2 * (obtener_de_compartida(fila - 1, (col / 4)) ? memBloque[indexBloque - blockDim.x * 4 + i] : imgIn[index - widthStep + i]) + (obtener_de_compartida(fila - 1, (col / 4) + 1 ) ? memBloque[indexBloque - blockDim.x * 4 + 4 + i] : imgIn[index - widthStep + 4 + i]) 	- (obtener_de_compartida(fila + 1, (col / 4) - 1 ) ? memBloque[indexBloque + blockDim.x * 4 - 4 + i] : imgIn[index + widthStep - 4 + i]) 	- 2 * (obtener_de_compartida(fila + 1, (col / 4)) ? memBloque[indexBloque + blockDim.x * 4 + i] : imgIn[index + widthStep + i]) - (obtener_de_compartida(fila + 1, (col / 4) + 1) ? memBloque[indexBloque + blockDim.x * 4 + 4 + i] : imgIn[index + widthStep + 4 + i]);
	
			if (resultado < 0) resultado = 0;
			else if (resultado >255) resultado = 255;			
			kernel0[i] = (unsigned char)resultado;

			resultado = 
	- (obtener_de_compartida(fila - 1, (col / 4) - 1 ) ? memBloque[indexBloque - blockDim.x * 4 - 4 + i] : imgIn[index - widthStep - 4 + i])
	+ (obtener_de_compartida(fila - 1, (col / 4) + 1 ) ? memBloque[indexBloque - blockDim.x * 4 + 4 + i] : imgIn[index - widthStep + 4 + i])
	- 2 * (obtener_de_compartida(fila, (col / 4) - 1 ) ? memBloque[indexBloque - 4 + i] : imgIn[index - 4 + i])
	+ 2 * (obtener_de_compartida(fila, (col / 4) + 1 ) ? memBloque[indexBloque + 4 + i] : imgIn[index + 4 + i])
	- (obtener_de_compartida(fila + 1, (col / 4) - 1 ) ? memBloque[indexBloque + blockDim.x * 4 - 4 + i] : imgIn[index + widthStep - 4 + i])
	+ (obtener_de_compartida(fila + 1, (col / 4) + 1 ) ? memBloque[indexBloque + blockDim.x * 4 + 4 + i] : imgIn[index + widthStep + 4 + i]);

			if (resultado < 0) resultado = 0;
			else if (resultado >255) resultado = 255;			
			kernel1[i] = (unsigned char)resultado;

			imgOut[index + i] = sqrtf( kernel0[i] * kernel0[i] + kernel1[i] * kernel1[i]);
		}
	}
}

extern "C" void sobelCUDA(unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStep, unsigned int height );

void sobelCUDA(unsigned char* imgIn, unsigned char* imgOut, unsigned int widthStep, unsigned int height ) 
{
	const int blockWidth = 16;	
	const int blockHeight = 16;

	unsigned char* in;
	unsigned char* out;
	const unsigned int size = widthStep * height * sizeof(unsigned char);
	
	cudaMalloc( (void**)&in, size );
	cudaMalloc( (void**)&out, size );
	cudaMemcpy( in, imgIn, size, cudaMemcpyHostToDevice );

	dim3 dimBlock( blockWidth, blockHeight );
	dim3 dimGrid( (widthStep / (4 * dimBlock.x))+1, (height / dimBlock.y) + 1 );

	sobel_matriz<<<dimGrid, dimBlock>>>( in, out, widthStep, height );

	cudaMemcpy( imgOut, out, size, cudaMemcpyDeviceToHost );
	cudaFree( in ); cudaFree( out );
}
